/*
 *	Copyright (c) 2005-2006 Applied Networking Inc. All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms set out in the LICENSE.tuncfg file, which is 
 *	included in Hamachi Client distribution.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	LICENSE.tuncfg file for more details.
 *
 *	(BSD license)
 */

/*
 *	Tuncfg is a tunneling device management and configuration daemon.
 *
 *	The idea is that tuncfg encapsulates all root-level functionality
 *	normally required by a private networking software. Namely -
 *
 *	* creation of tunneling devices; this requires an access to 
 *	  /dev/tapXX files, which _usually_ has 700 access mask
 *
 *	* configuration of the tunneling device using ifconfig, which is
 *	  always a root-level operation
 *
 *	Tuncfg must be run with superuser privileges. It will daemonize and
 *	open a listening domain socket /var/run/tuncfg.sock.
 *
 *	Upon accepting the connection on this socket, it will issue an open()
 *	call for /dev/tapXX file (iterating over first 16 XX values)
 *	and pass obtained FD to the peer process. It will also query and pass
 *	the MAC address of the device to the peer process.
 *
 *	It will then listen for ifconfig requests from the peer. The request 
 *	contains just an IP address and mask. Tuncfg will lookup the tun 
 *	device the peer owns and will assign IP/mask to it using system() 
 *	call for ipconfig command.
 *
 *	'tuncfg'     starts the daemon (it won't start twice, there's a check)
 *	'tuncfg -d'  starts the console version (useful for debugging)
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <arpa/inet.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <netinet/in_var.h>

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

/*
 *
 */
#define LOCK_PATH   "/var/run/tuncfg.lock"
#define SOCK_PATH   "/var/run/tuncfg.sock"

#define TUNTAP_URL  "http://www.hamachi.cc/tuntap"

#define MAX_CLIENTS 16

struct context
{
	int   fd;
	char  dev[IFNAMSIZ];
};

/*
 *
 */
#define errorf(args...) { printf(args); exit(1); }
 
int send_with_fd(int fd, const void * buf, size_t len, int fd_to_send);
int darwin_fixup(const char * dev, unsigned long addr, unsigned long mask);

int main(int argc, char ** argv)
{
	struct sockaddr_un addr = { 0 };
	socklen_t alen = sizeof(addr);
	struct stat st;
	pid_t pid;
	int   fd, r, i;
	int   debug = 0;

	struct context ctx[MAX_CLIENTS];
	int ctx_n = 0;

	// check we are root
	if (getuid() != 0)
		errorf("tuncfg: must be run with superuser permissions\n");

	//
	if (argc > 1)
	{
		debug = (strcmp(argv[1], "-d") == 0);
	}

	// lock
	fd = open(LOCK_PATH, O_CREAT | O_RDWR);
	if (fd < 0)
		errorf("tuncfg: cannot open lock file %s -- %s\n",
			LOCK_PATH, strerror(errno));

	//
	if (flock(fd, LOCK_EX | LOCK_NB) < 0)
	{
		errorf("tuncfg: already running, "
		       "use 'killall tuncfg; tuncfg' to restart it\n");
	}

	// check there's /dev/net/tun
	if (stat("/dev/tap0", &st) < 0)
		errorf("tuncfg: cannot stat() /dev/tap0 -- %s\n"
		       "tuncfg: visit %s for more information\n",
		       strerror(errno), TUNTAP_URL);
	
	// open request socket
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, SOCK_PATH);

	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
		errorf("tuncfg: cannot create domain socket -- %s\n",
			strerror(errno));
	
	unlink(SOCK_PATH);
	r = bind(fd, (void*)&addr, alen);
	if (r < 0)
		errorf("tuncfg: cannot bind domain socket to %s -- %s\n", 
			SOCK_PATH, strerror(errno));

	r = listen(fd, 2);
	if (r < 0)
		errorf("tuncfg: cannot listen on domain socket %s -- %s\n",
			SOCK_PATH, strerror(errno));

	// make the socket world-accessible
	if (stat(SOCK_PATH, &st) < 0)
		errorf("tuncfg: cannot stat() %s -- %s\n",
			SOCK_PATH, strerror(errno));

	if (chmod(SOCK_PATH, st.st_mode | 0777) < 0)
		printf("tuncfg: failed to chmod +777 %s -- %s\n",
			SOCK_PATH, strerror(errno));

	// daemonize
	if (! debug)
	{
		chdir("/");

		pid = fork();
		if (pid < 0) return 0;
		if (pid > 0) exit(0);
		
		if (setsid() < 0)
			return 0;/* shouldn't happen */

		pid = fork();
		if (pid < 0) return 0;
		if (pid > 0) exit(0);

		close(0);
		close(1);
		close(2);

		open("/dev/null", O_RDWR);
		dup(0);
		dup(0);
	}

	// loop
	for (;;)
	{
		fd_set fdr;
		int    fdm;

		FD_ZERO(&fdr);

		FD_SET(fd, &fdr);
		fdm = fd;

		for (i=0; i<ctx_n; i++)
		{
			FD_SET(ctx[i].fd, &fdr);
			if (ctx[i].fd > fdm)
				fdm = ctx[i].fd;
		}

		do
		{
			r = select(fdm+1, &fdr, NULL, NULL, NULL);
			if (r < 0 && errno != EINTR)
				exit(EXIT_FAILURE);
		}
		while (r < 0);

		if (FD_ISSET(fd, &fdr))
		{
			struct context * p;
			char buf[4+6];
			int mib[6];
			size_t len;
			struct if_msghdr * msg = NULL;
			struct sockaddr_dl * sa;
			char dev_name[32];
			int i;
			int cli, dev = -1, tmp = -1;

			cli = accept(fd, (void*)&addr, &alen);
			printf("tuncfg: accept() %d %d\n", cli, errno);
			
			if (cli < 0)
				goto skip;

			if (ctx_n == MAX_CLIENTS)
			{
				printf("tuncfg: too many clients\n");
				r = (0x01 << 24); 
				goto done;
			}

			// open first available tap device
			for (i=0; i<MAX_CLIENTS; i++)
			{
				snprintf(dev_name, sizeof(dev_name),
						"/dev/tap%d", i);

				dev = open(dev_name, O_RDWR);
				printf("tuncfg: open(%s) %d %d\n", 
						dev_name, dev, errno);
				if (dev >= 0)
					break;
			}
			if (dev < 0)
			{
				r = (0x02 << 24);
				goto done;
			}

			// query mac
			mib[0] = CTL_NET;
			mib[1] = AF_ROUTE;
			mib[2] = 0;
			mib[3] = AF_LINK;
			mib[4] = NET_RT_IFLIST;
			mib[5] = if_nametoindex("tap0");

			if (! mib[5])
			{
				r = (0x03 << 24) | errno;
				goto done;
			}

			if (sysctl(mib, 6, NULL, &len, NULL, 0) < 0)
			{
				r = (0x04 << 24) | errno;
				goto done;
			}

			msg = malloc(len);
			if (! msg)
			{
				r = (0x05 << 24) | errno;
				goto done;
			}

			if (sysctl(mib, 6, msg, &len, NULL, 0) < 0)
			{
				r = (0x05 << 24) | errno;
				goto done;
			}

			sa = (void*)(msg + 1);
			memcpy(buf+4, LLADDR(sa), 6);

			// remember
			p = &ctx[ctx_n++];
			p->fd = cli;
			strncpy(p->dev, dev_name+5, 5);

			r = 0;
done:
			if (cli != -1)
			{
				memcpy(buf, &r, sizeof(r));
				send_with_fd(cli, buf, sizeof(buf), dev); 
			}

			free(msg);
			if (tmp != -1)  close(tmp);
			if (dev != -1)  close(dev);
			if (r != 0)     close(cli);
		}

skip:
		for (i=0; i<ctx_n; i++)
		{
			unsigned char v[8];
			unsigned long a, m, b;
			char cmd[256];
			
			if (! FD_ISSET(ctx[i].fd, &fdr))
				continue;
		
			r = recv(ctx[i].fd, v, sizeof(v), 0);
			printf("tuncfg: recv() %d %d\n", r, errno);
			if (r == 0)
			{
				/* client's gone */
				printf("tuncfg: removing %s\n", ctx[i].dev);
				close(ctx[i].fd);
				ctx[i--] = ctx[ctx_n--];
				continue;
			}
			
			if (r != sizeof(v))
			{
				r = (0x06 << 24);
				goto ack;
			}

			/* v[0-3] = ip, v[4-7] = mask */
			a = v[0];
			a = (a << 8) | v[1];
			a = (a << 8) | v[2];
			a = (a << 8) | v[3];

			m = v[4];
			m = (m << 8) | v[5];
			m = (m << 8) | v[6];
			m = (m << 8) | v[7];

			if ( (a & 0xff000000) != 0x05000000 ||
			     (m & 0xff000000) != 0xff000000 )
			{
				printf("tuncfg: bad arg %08x %08x\n", a, m);

				r = (0x07 << 24);
				goto ack;
			}
			
			r  = sprintf(cmd, 
				"ifconfig %s %u.%u.%u.%u ", ctx[i].dev,
				v[0], v[1], v[2], v[3]);
		
			r += sprintf(cmd + r, 
				"netmask %u.%u.%u.%u ",
				v[4], v[5], v[6], v[7]);

			b = a | ~m;
			r += sprintf(cmd + r,
				"mtu 1200 broadcast %u.%u.%u.%u",
				(b >> 24) & 0xff, (b >> 16) & 0xff,
				(b >>  8) & 0xff, (b >>  0) & 0xff);

			r = system(cmd);
			printf("tuncfg: system(%s) %d %d\n", cmd, r, errno);
			if (r != 0)
			{
				r = (0x08 << 24) | (r & 0x00ffffff);
				goto ack;
			}

			if (darwin_fixup(ctx[i].dev, a, m) < 0)
			{
				r = (0x09 << 24);
				goto ack;
			}
ack:
			printf("tuncfg: config() %08x\n", r);

			v[0] = 0xff & (r >> 24);
			v[1] = 0xff & (r >> 16);
			v[2] = 0xff & (r >> 8);
			v[3] = 0xff & r;
			send_with_fd(ctx[i].fd, v, 4, -1);
		}
	}

	return 0;
}

int send_with_fd(int fd, const void * ptr, size_t len, int fd_to_send)
{
	struct msghdr    msg = { 0 };
	struct iovec     iov[1];
	struct cmsghdr * hdr;
	char buf[CMSG_SPACE(sizeof(int))];

	if (fd_to_send != -1)
	{
		msg.msg_control = buf;
		msg.msg_controllen = sizeof(buf);

		hdr = CMSG_FIRSTHDR(&msg);
		hdr->cmsg_len = CMSG_LEN(sizeof(int));
		hdr->cmsg_level = SOL_SOCKET;
		hdr->cmsg_type = SCM_RIGHTS;
		*(int*)CMSG_DATA(hdr) = fd_to_send;
	}

	iov[0].iov_base = (void*)ptr;
	iov[0].iov_len  = len;
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;

	return sendmsg(fd, &msg, 0);
}

int darwin_fixup(const char * dev, unsigned long addr, unsigned long mask)
{
	struct in_aliasreq req = { 0 };
	unsigned long temp;
	int fd, ret;

	/*
	 *	Re-assign a netmask so that xnu copies ifra_mask into 
	 *	its own data and thus (re-)initializes sin_family field 
	 *	in its copy of netmask's sockaddr. mDNSresolver checks
	 *	this field, so it must be in order for Bonjour to work
	 *	over this interface.
	 */
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
	{
		printf("tuncfg: darwin_fixup() failed\n");
		return -1;
	}

	temp = (mask == 0xffffffff) ? 0xffffff00 : 0xffffffff;

	strcpy(req.ifra_name, dev);

	req.ifra_addr.sin_family = AF_INET;
	req.ifra_addr.sin_len = sizeof(struct sockaddr_in);
	req.ifra_addr.sin_addr.s_addr = htonl(addr);

	req.ifra_mask.sin_family = AF_INET;
	req.ifra_mask.sin_len = sizeof(struct sockaddr_in);

	//
	req.ifra_mask.sin_addr.s_addr = htonl(temp);
	ret = ioctl(fd, SIOCAIFADDR, &req);
	if (ret < 0)
	{
		printf("tuncfg: ioctl(SIOCAIFADDR) failed %d\n", errno);
		goto err;
	}
	
	//
	req.ifra_mask.sin_addr.s_addr = htonl(mask);
	ret = ioctl(fd, SIOCAIFADDR, &req);
	if (ret < 0)
	{
		printf("tuncfg: ioctl(SIOCAIFADDR, 2) failed %d\n", errno);
		goto err;
	}
err:
	close(fd);
	return ret;

}

